from polars import DataFrame as POLARS_DF
from .interface import (
    SourceInterface,
    SourceInterfaceConfiguration,
    SourceInterfaceConnectionSettings,
    register_source,
)


class FakerSourceConfiguration(SourceInterfaceConfiguration):
    pass


class FakerSourceConnectionSettings(SourceInterfaceConnectionSettings):
    """The connection configuration class used for faker sources"""

    data: POLARS_DF

    def __init__(self, faker: dict):
        self.data = POLARS_DF(faker)


@register_source("faker", FakerSourceConfiguration, FakerSourceConnectionSettings)
class FakerSource(SourceInterface):
    ConnectionSettingsClass = FakerSourceConnectionSettings
    SourceConfigClass = FakerSourceConfiguration

    """ A source for faker data """

    def _configure(self):
        raise Exception("Not yet implemented")

    def _connect(self):
        pass

    def _disconnect(self):
        pass

    def _query(self, params=None) -> POLARS_DF:
        return self.connection_settings.data

    def insert(self, data: POLARS_DF):
        print("Inserting data into static source...")
        print(data)
        return len(data)

    def update(self, data: POLARS_DF):
        print("Updating data in static source...")
        print(data)
        return len(data)

    def delete(self, data: POLARS_DF):
        print("Deleting data from static source")
        print(data)
        return len(data)



"""Module version information."""

import os
import re

REGULAR_RELEASE = "(\d.\d.\d)"
PRE_RELEASE = "(\d.\d.\d-rc.\d)"

gh_tag = os.getenv("GHTAG")
gh_branch = os.getenv("GHBRANCH")
gh_run = os.getenv("GHRUN")

__version__ = "0.0.1"

if gh_tag is not None:
    matches = re.findall(PRE_RELEASE, gh_tag)
    if matches and len(matches) == 1:
        __version__ = matches[0]
    else:
        matches = re.findall(REGULAR_RELEASE, gh_tag)
        if matches and len(matches) == 1:
            __version__ = matches[0]



import polars as pl
import sqlalchemy as sqla
from typing import List
from .interface import (
    register_source,
    SourceInterface,
    SourceInterfaceConfiguration,
    SourceInterfaceConnectionSettings,
)


class MysqlSourceConfiguration(SourceInterfaceConfiguration):
    """The configuration class used for MySQL sources"""

    unique_columns: List[str] = None
    skip_columns: List[str] = None
    table: str = None
    query: str = None

    def __init__(self, table: str = None, query: str = None, uniqueColumns: List[str] = [], skipColumns: List[str] = []):
        super().__init__()
        self.table = table
        self.query = query
        self.unique_columns = uniqueColumns
        self.skip_columns = skipColumns


class MysqlSourceConnectionSettings(SourceInterfaceConnectionSettings):
    """The connection configuration class used for MySQL sources"""

    connection_string: str
    query: str = None
    table: str = None

    def __init__(self, settings: dict):
        if settings.get("connection_string", False):
            self.connection_string = settings["connection_string"]
            return

        self.connection_string = "mysql+pymysql://"
        f"{settings['username']}:{settings['password']}"
        f"@{settings['host']}:{settings['port']}/{settings['database']}"


@register_source("mysql", MysqlSourceConfiguration, MysqlSourceConnectionSettings)
class MysqlSource(SourceInterface):
    ConnectionSettingsClass = MysqlSourceConnectionSettings
    SourceConfigClass = MysqlSourceConfiguration

    """ A source for MySQL data """

    def _configure(self):
        pass

    def _connect(self):
        pass

    def _disconnect(self):
        pass

    def _query(
        self, params=None, customQuery: str = None, returnData: bool = True
    ) -> pl.DataFrame:
        query = customQuery

        if query is None:
            if self.source_configuration.query is not None:
                query = self.source_configuration.query

            if query is None:
                if self.source_configuration.table is None:
                    raise Exception("No query or table specified")

                query = f"SELECT * FROM {self.source_configuration.table}"

        if returnData:
            try:
                df = pl.read_database_uri(
                    query=query,
                    uri=self.connection_settings.connection_string,
                )
            except ModuleNotFoundError:
                df = pl.read_database_uri(
                    query=query,
                    uri=self.connection_settings.connection_string.replace(
                        "mysql://", "mysql+pymysql://"
                    ),
                )

            return df
        try:
            with sqla.create_engine(
                self.connection_settings.connection_string
            ).connect() as con:
                con.execute(query)
                con.commit()

        except ModuleNotFoundError:
            with sqla.create_engine(
                self.connection_settings.connection_string.replace(
                    "mysql://", "mysql+pymysql://"
                )
            ).connect() as con:
                con.execute(sqla.text(query))
                con.commit()

    def _insert(
        self, data: pl.DataFrame, table: str = None, connection_string: str = None
    ):
        if table is None:
            table = self.source_configuration.table
        if connection_string is None:
            connection_string = self.connection_settings.connection_string

        try:
            data.write_database(
                table, self.connection_settings.connection_string, if_table_exists="append"
            )
        except ModuleNotFoundError:
            data.write_database(
                table,
                self.connection_settings.connection_string.replace(
                    "mysql://", "mysql+pymysql://"
                ),
                if_table_exists="append",
            )

        return len(data)

    def insert(self, data: pl.DataFrame):
        self._validate_unique_columns()
        return self._insert(data)

    def update(self, data: pl.DataFrame):
        self._validate_unique_columns()

        tempDB = self.source_configuration.table + "_udTemp"
        try:
            with sqla.create_engine(
                self.connection_settings.connection_string
            ).connect() as con:
                con.execute(sqla.text(f"DROP TABLE IF EXISTS {tempDB}"))
                con.execute(
                    sqla.text(
                        f"CREATE TEMPORARY TABLE {tempDB} "
                        f"LIKE {self.source_configuration.table}"
                    )
                )
                con.commit()

        except ModuleNotFoundError:
            with sqla.create_engine(
                self.connection_settings.connection_string.replace(
                    "mysql://", "mysql+pymysql://"
                )
            ).connect() as con:
                con.execute(sqla.text(f"DROP TABLE IF EXISTS {tempDB}"))
                con.execute(
                    sqla.text(
                        f"CREATE TEMPORARY TABLE {tempDB} "
                        f"LIKE {self.source_configuration.table}"
                    )
                )
                con.commit()

        # Insert data into temporary table
        self._insert(data, table=tempDB)

        fields_to_update = [field.name for field in data.get_columns(
        ) if not field.name in self.source_configuration.skip_columns or not field.name in self.source_configuration.unique_columns]

        field_spec = ", ".join(
            (f"`{fieldName}`" for fieldName in fields_to_update))

        query = f"""
            REPLACE INTO {self.source_configuration.table} ({field_spec})
            SELECT {field_spec} FROM {tempDB}
        """

        self._query(customQuery=query, returnData=False)

        return len(data)

    def delete(self, data: pl.DataFrame):
        batch_size = 500
        batches = [data]

        if len(data) > batch_size:
            batches = []

            for i in range(0, len(data), batch_size):
                batches.append(data[i: i + batch_size])

        for batch in batches:
            id_clause = " AND ".join(
                (
                    f"`{fieldName}` IN ({','.join([self._quote_if_needed(x) for x in batch[fieldName].to_list()])})"
                    for fieldName in self.source_configuration.unique_columns
                )
            )

            query = f"""
                DELETE FROM {self.source_configuration.table}
                WHERE {id_clause}
            """

            self._query(customQuery=query, returnData=False)

        return len(data)

    def _validate_unique_columns(self):
        if not self.source_configuration.unique_columns:
            raise ValueError(
                "MySQL source requires the unique_columns to be set if used as a destination")

    def _quote_if_needed(self, id: any) -> str:
        if isinstance(id, str):
            return f"'{id}'"
        return str(id)



from typing import Any, Dict
import polars as pl
from .interface import TransformerInterface, register_transformer_class
import json


@register_transformer_class("structs")
class StructTransformers(TransformerInterface):
    @staticmethod
    def staticfield(data: pl.DataFrame, field: str, value):
        """Add a struct field to the DataFrame

        Args:
            data (pl.DataFrame): The dataFrame to modify
            field (str): The field to add
            value (str): The value of the field to add

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        data = data.with_columns(pl.Series(field, [value] * len(data)))
        return data

    @staticmethod
    def jsonpath(
        data: pl.DataFrame,
        inField: str,
        outField: str,
        jsonPath: str,
        defaultValue: str = "",
    ):
        """
        Retrieve a value from a jsonpath inside of a column and add it to a new column

        Args:
            data (pl.DataFrame): The dataFrame to modify
            field (str): The field to extract the jsonpath from
            outField (str): The field to add the extracted value to
            jsonPath (str): The jsonpath to extract
        """
        split_jsonpath = jsonPath.split(".")

        if len(split_jsonpath) == 1 and "$" in split_jsonpath:
            return data.with_columns(pl.Series(outField, data[inField]))

        def get_value_at_jsonpath(fData: Any, start_at: int = 0):
            if isinstance(fData, str):
                fData = json.loads(fData)

            try:
                for index, selector in enumerate(
                    split_jsonpath[start_at:], start=start_at
                ):
                    if selector == "$":
                        continue
                    if selector == "*":
                        if not isinstance(fData, list) and not isinstance(
                            fData, pl.Series
                        ):
                            raise NotImplementedError(
                                "Wildcard selector is only supported for lists"
                            )
                        fData = [
                            get_value_at_jsonpath(fData[j], index + 1)
                            for j in range(len(fData))
                        ]
                        return fData
                    try:
                        ix = int(selector)
                        fData = fData[ix]
                    except:
                        fData = fData[selector]
                return fData
            except:
                return defaultValue

        dCol = data[inField]
        outCol = []
        for i in range(len(dCol)):
            outCol.append(get_value_at_jsonpath(dCol[i]))

        data = data.with_columns(pl.Series(outField, outCol))

        return data

    @staticmethod
    def compose_struct(data: pl.DataFrame, map: Dict[str, str], outField: str):
        """Compose a struct from other columns as a new column in the DataFrame

        Given a DataFrame with the columns `age` containing `30` and `name` containing `John`. Providing the map `{"name": "name", "age", "age"}` with outField `person` will result in a new column `person` with the value `{"name": "John", "age": 30}`

        Args:
            data (pl.DataFrame): The dataFrame to modify
            map (dict): Dict of fields to compose the struct from, {"name_of_field_in_struct": "name_of_column"}.
            outField (str): Name of the column to add

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        fields = list(map.keys())
        __class__._validate_fields(data.columns, fields)

        rows = len(data)
        newStructs = []
        for i in range(rows):
            newStructs.append({name: data[field][i] for name, field in map.items()})

        series = pl.Series(outField, newStructs)
        data = data.with_columns(series)

        return data

    @staticmethod
    def compose_list_of_structs(data: pl.DataFrame, map: Dict[str, str], outField: str):
        """Compose a list of structs from other columns as a new column in the DataFrame

        All source columns must be lists. If the lists are of different lengths, the resulting list will be the length of the longest list with None as the value for missing elements.
        table

        Args:
            data (pl.DataFrame): The dataFrame to modify.
            map (dict): Dict of fields to compose the struct from, {"name_of_field_in_struct": "name_of_column"}.
            outField (str): Name of the column to add.

        Returns:
            pl.DataFrame: The resulting DataFrame.

        Example:
            The source dataframe

            +----+----------------------------+----------+----------+
            | id | names                      | ages     | city     |
            +----+----------------------------+----------+----------+
            | 1  | ["John", "Mary", "Sam"]    | [30, 40] | New York |
            +----+----------------------------+----------+----------+

            Transformed with this configuration

            `compose_list_of_structs(data, {"name": "names", "age": "ages"}, "people")`

            Will result in a people column composed of the folowing list

            [
                {"name": "John", "age": 30},
                {"name": "Mary", "age": 40},
                {"name": "Sam", "age": None}
            ]
        """
        field_names = list(map.values())
        __class__._validate_fields(data.columns, field_names)

        dataframe_to_map = data.select(field_names)

        longest_series_row = 0
        for series in dataframe_to_map:
            for row in series:
                if len(row) > longest_series_row:
                    longest_series_row = len(row)

        new_series = None
        row_count = len(data)
        for row_index in range(row_count):
            new_structs = []
            for i in range(longest_series_row):
                new_struct = {}
                for name, series in map.items():
                    try:
                        new_struct[name] = dataframe_to_map[series][row_index][i]
                    except IndexError:
                        new_struct[name] = None

                new_structs.append(new_struct)
            if new_series is None:
                new_series = pl.Series([new_structs])
                continue
            new_series.append(pl.Series([new_structs]))

        data = data.with_columns(new_series.alias(outField))

        return data



from typing import List
from polars import DataFrame as POLARS_DF
from .interface import (
    register_source,
    SourceInterface,
    SourceInterfaceConfiguration,
    SourceInterfaceConnectionSettings,
)


class StaticSourceConfiguration(SourceInterfaceConfiguration):
    """The configuration class used for static sources"""

    pass


class StaticSourceConnectionSettings(SourceInterfaceConnectionSettings):
    """The connection configuration class used for static sources"""

    data: POLARS_DF

    def __init__(self, static: dict):
        self.data = POLARS_DF(static)


@register_source("static", StaticSourceConfiguration, StaticSourceConnectionSettings)
class StaticSource(SourceInterface):
    ConnectionSettingsClass = StaticSourceConnectionSettings
    SourceConfigClass = StaticSourceConfiguration

    """ A source for static data """

    def _configure(self):
        pass

    def _connect(self):
        pass

    def _disconnect(self):
        pass

    def _query(self, params=None) -> POLARS_DF:
        return self.connection_settings.data

    def insert(self, data: POLARS_DF):
        print("Inserting data into static source...")
        print(data)
        return len(data)

    def update(self, data: POLARS_DF):
        print("Updating data in static source...")
        print(data)
        return len(data)

    def delete(self, data: POLARS_DF):
        print("Deleting data from static source")
        print(data)
        return len(data)



import os

cDir = os.path.dirname(__file__)
for file in os.listdir(cDir):
    if file.endswith(".py") and not file.startswith("__") and file != "interface.py":
        __import__(f"{__name__}.{file[:-3]}", fromlist=["*"])



from enum import Enum
from typing import Union
from polars import DataFrame as POLARS_DF

FUNC_TYPE = type(print)
CLS_TYPE = type(Enum)


def register_transformer(namespace: str, name: str):
    def wrapper(func: FUNC_TYPE):
        Transformers._registerFunction(namespace, name, func)
        return func

    return wrapper


def register_transformer_class(namespace: str):
    def wrapper(cls: CLS_TYPE):
        Transformers._registerClass(cls, namespace)
        return cls

    return wrapper


class TransformerInterface:
    @staticmethod
    def _validate_fields(
        columns: Union[list, set, tuple], fields: Union[str, list, set, tuple]
    ):
        if isinstance(fields, str):
            fields = [fields]

        if sum([1 for col in columns if col in fields]) == len(columns):
            raise KeyError(
                f'Some of the field(s) {",".join(fields)} are not present '
                "in the dataset. Valid columns at this stage are:"
                f'{",".join(columns)}'
            )
        return


class Transformers:
    transformers: dict = {}

    @staticmethod
    def _registerFunction(namespace: str, name: str, func: FUNC_TYPE):
        Transformers.transformers[f"{namespace}.{name}"] = func

    @staticmethod
    def _registerClass(cls: CLS_TYPE, namespace: str = None):
        namespace = namespace if namespace is not None else cls.__name__
        for fun in dir(cls):
            if not fun.startswith("_"):
                Transformers._registerFunction(namespace, fun, getattr(cls, fun))

    @staticmethod
    def runTransformer(transformer: str, data: POLARS_DF, **kwargs) -> POLARS_DF:
        try:
            return Transformers.transformers[transformer](data, **kwargs)
        except TypeError as e:
            raise TypeError(
                f"The wrong arguments supplied for transformer {transformer}: {str(e)}"
            ) from e
        except KeyError as e:
            raise KeyError(
                f"The transformer {transformer} does not exist or an error occured in the transformer: ",
                str(e),
            ) from e
        except Exception as e:
            raise Exception(
                f"An error occurred while running transformer {transformer}: {str(e)}"
            ) from e


class TransformerConfiguration:
    """The configuration class for transformers"""

    identifier: str
    config: dict = None
    include_sync: bool = None

    def __init__(
        self, identifier: str, config: dict = None, include_sync: bool = False
    ) -> None:
        self.identifier = identifier
        self.config = config
        self.include_sync = include_sync

    def transform(self, data: POLARS_DF, **kwargs):
        config = self.config or {}
        if data.is_empty():
            return data

        if len(data) == 0:
            return data

        return Transformers.runTransformer(self.identifier, data, **config, **kwargs)



#!/usr/bin/env python3
import os
from setuptools import setup

repository_name = "beetl"
module_name = "beetl"

# Get key package details
version = {}  # type: ignore

# Load the __version__.py file from the plugin directory
with open(os.path.join(os.getcwd(), "src", module_name, "__version__.py")) as f:
    exec(f.read(), version)

# Load the README file and use it as the long_description for PyPI
with open("README.md", "r") as f:
    readme = f.read()

# Package configuration - for reference see:
# https://setuptools.readthedocs.io/en/latest/setuptools.html#id9
setup(
    long_description=readme,
    long_description_content_type="text/markdown",
    version=version["__version__"],
)



import os
import json

# import asyncio
import urllib3
import polars as pl
import requests
import requests.adapters
from alive_progress import alive_bar
from .interface import (
    register_source,
    SourceInterface,
    SourceInterfaceConfiguration,
    SourceInterfaceConnectionSettings,
    RequestThreader,
)

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

BULK_CUTOFF = 300

DATAMODELS_WITHOUT_SOFT_DELETE = (
    "NutanixCluster",
    "NutanixClusterHost",
    "NutanixVM",
    "NutanixVMDisk",
    "NutanixNetwork",
    "NutanixNetworkInterface",
)


class ItopSourceConfiguration(SourceInterfaceConfiguration):
    """The configuration class used for iTop sources"""

    datamodel: str = None
    oql_key: str = None
    soft_delete: dict = None
    link_columns: list = None
    comparison_columns: list = None
    unique_columns: list = None
    skip_columns: list = None

    def __init__(
        self,
        datamodel: str = None,
        oql_key: str = None,
        soft_delete: dict = None,
        comparison_columns: list = None,
        unique_columns: list = None,
        link_columns: list = [],
        skip_columns: list = None,
    ):
        super().__init__()
        self.datamodel = datamodel
        self.oql_key = oql_key

        if comparison_columns is None:
            raise Exception("Comparison columns must be provided")
        self.comparison_columns = comparison_columns

        if unique_columns is None:
            raise Exception("Unique columns must be provided")
        self.unique_columns = unique_columns

        self.skip_columns = skip_columns

        if soft_delete is not None:
            if (
                soft_delete.get("enabled", False)
                and datamodel in DATAMODELS_WITHOUT_SOFT_DELETE
            ):
                raise Exception(
                    f"Soft delete is not supported for {datamodel} datamodel"
                )

        self.soft_delete = soft_delete

        self.link_columns = link_columns


class ItopSourceConnectionSettings(SourceInterfaceConnectionSettings):
    """The connection configuration class used for iTop sources"""

    host: str
    username: str = None
    password: str = None
    verify_ssl: bool = True

    def __init__(self, settings: dict):
        self.host = settings.get("host", False)
        self.username = settings.get("username", False)
        self.password = settings.get("password", False)
        self.verify_ssl = settings.get("verify_ssl", "true") == "true"


@register_source("itop", ItopSourceConfiguration, ItopSourceConnectionSettings)
class ItopSource(SourceInterface):
    ConnectionSettingsClass = ItopSourceConnectionSettings
    SourceConfigClass = ItopSourceConfiguration
    auth_data: dict = None

    """ A source for Combodo iTop Data data """

    def soft_delete_active(self) -> bool:
        """Check if the soft_delete option is active

        Returns:
            bool: True/False
        """
        if self.source_configuration.soft_delete is None:
            return False

        return self.source_configuration.soft_delete.get("enabled", False)

    def _configure(self):
        self.auth_data = {
            "auth_user": self.connection_settings.username,
            "auth_pwd": self.connection_settings.password,
        }

        self.url = f"https://{self.connection_settings.host}/webservices/rest.php?version=1.3&login_mode=form"

        # Test Connection
        files = self._create_body(
            "core/check_credentials",
            {
                "user": self.connection_settings.username,
                "password": self.connection_settings.password,
            },
        )

        res = requests.post(
            self.url,
            files=files,
            data=self.auth_data,
            verify=self.connection_settings.verify_ssl,
        )

        try:
            res_json = res.json()
            assert res_json["code"] == 0
        except AssertionError:
            raise Exception(
                f"Invalid response received from iTop Server: {res_json['code']}"
            )
        except requests.exceptions.JSONDecodeError:
            raise Exception(
                "Invalid response received from iTop Server, "
                "check your authentication settings: " + res.text
            )

    def _create_body(self, operation: str, params: dict = None):
        data = params or {}

        data["operation"] = operation

        return {"json_data": json.dumps(data)}

    def _connect(self):
        pass

    def _disconnect(self):
        pass

    def _query(self, params=None) -> pl.DataFrame:
        """
        In:
            Object: The type of object to retrieve from iTop
            OQL: OQL Query
            Output Fields: List of fields to return
        """
        oql = self.source_configuration.oql_key
        if self.soft_delete_active():
            soft_delete = self.source_configuration.soft_delete
            if f"{soft_delete.get('field', 'status').lower()}" not in oql.lower():
                combining_word = "WHERE" if "WHERE" not in oql.upper() else "AND"
                oql += (
                    f" {combining_word} {soft_delete.get('field', 'status')} "
                    + f"= '{soft_delete.get('active_value', 'enabled')}'"
                )

        all_colums = (
            self.source_configuration.unique_columns
            + self.source_configuration.comparison_columns
            + self.source_configuration.link_columns
        )
        files = self._create_body(
            "core/get",
            {
                "class": self.source_configuration.datamodel,
                "key": oql,
                "output_fields": ",".join(all_colums),
            },
        )

        res = requests.post(
            self.url,
            files=files,
            data=self.auth_data,
            verify=self.connection_settings.verify_ssl,
        )

        if not res.ok or res.json()["code"] != 0:
            raise Exception(
                f"Error retrieving {self.source_configuration.datamodel}"
                f"data from iTop: {res.text}"
            )

        re_obj = []

        try:
            res_json = res.json()
            objects = res_json.get("objects", None)
            if objects is None:
                return pl.DataFrame()

            for _, item in res_json.get("objects", {}).items():
                if "id" in all_colums:
                    re_obj.append({"id": item["key"], **item["fields"]})
                elif "key" in all_colums:
                    re_obj.append({"key": item["key"], **item["fields"]})
                else:
                    re_obj.append(item["fields"])

            return pl.DataFrame(re_obj)

        except Exception as e:
            raise Exception(
                "An error occured while trying to decode the response", e
            ) from e

    def create_item(self, type: str, data: dict, comment: str = ""):

        body = self._create_body(
            "core/create",
            {
                "comment": comment,
                "class": type,
                "output_fields": "*",
                "fields": {
                    x: y
                    for x, y in data.items()
                    if y is not None and x not in self.source_configuration.link_columns
                },
            },
        )

        res = requests.post(
            self.url,
            files=body,
            data=self.auth_data,
            verify=self.connection_settings.verify_ssl,
        )

        if not res.ok:
            print("Not created: ", res.text, body)
            return False

        try:
            res_json = res.json()
            assert res_json["code"] == 0
        except Exception as e:
            print("Not Created", e, res.text, body)
            print(res.text + str(e))
            return False

        return True

    def update_item(self, type: str, data: dict, comment: str = ""):
        identifiers = {}

        for column_name in tuple(data.keys()):
            if column_name in self.source_configuration.unique_columns:
                identifiers[column_name] = data.pop(column_name)

        # if len(identifiers) == 0:
        #    dKeys = tuple(data.keys())
        #    for column_name in self.source_configuration.unique_columns:
        #        if column_name in dKeys:
        #            identifiers[column_name] = data.pop(column_name)

        where_clause_conditions = " AND ".join(
            [f"{key} = '{value}'" for key, value in identifiers.items()]
        )

        body = self._create_body(
            "core/update",
            {
                "comment": comment,
                "class": type,
                "key": f"SELECT {type} WHERE {where_clause_conditions}",
                "fields": {
                    key: value for key, value in data.items() if value is not None
                },
            },
        )

        res = requests.post(
            self.url,
            files=body,
            data=self.auth_data,
            verify=self.connection_settings.verify_ssl,
        )

        if not res.ok:
            print("Not updated: ", res.text, body)
            return False

        try:
            res_json = res.json()
            assert res_json["code"] == 0
        except Exception as e:
            print("Not Updated", e, res.text, body)
            return False

        return True

    def delete_item(self, type: str, data: dict, comment: str = ""):
        identifiers = {}

        for key in tuple(data.keys()):
            if key in self.source_configuration.unique_columns:
                identifiers[key] = data.pop(key)

        where_clause_conditions = " AND ".join(
            [f"{key} = '{value}'" for key, value in identifiers.items()]
        )

        body = self._create_body(
            "core/delete",
            {
                "comment": comment,
                "class": type,
                "key": f"SELECT {type} WHERE {where_clause_conditions}",
            },
        )

        res = requests.post(
            self.url,
            files=body,
            data=self.auth_data,
            verify=self.connection_settings.verify_ssl,
        )

        if not res.ok:
            print("Not deleted: ", res.text, body)
            return False

        try:
            res_json = res.json()
            assert res_json["code"] == 0
        except Exception as e:
            print("Not Deleted", e, res.text, body)
            return False

        return True

    def insert(self, data: pl.DataFrame):
        insertData = data.clone()

        insert_cols = tuple(
            self.source_configuration.unique_columns
            + self.source_configuration.comparison_columns
        )

        for column in data.columns:
            if column not in insert_cols:
                insertData.drop_in_place(column)

        iters = (
            {
                "type": self.source_configuration.datamodel,
                "data": x,
                "comment": "Created via API Sync",
            }
            for x in insertData.iter_rows(named=True)
        )

        with alive_bar(len(data)) as pr_bar:
            if len(data) > BULK_CUTOFF:
                with RequestThreader(50) as needle:
                    for _ in needle.submitAndWait(self.create_item, iters):
                        pr_bar()
            else:
                for item in iters:
                    self.create_item(**item)
                    pr_bar()

        return len(data)

    def update(self, data: pl.DataFrame):
        updateData = data.clone()

        update_cols = tuple(
            column_name
            for column_name in self.source_configuration.unique_columns
            + self.source_configuration.comparison_columns
            if column_name not in (self.source_configuration.skip_columns or [])
        )

        for column in data.columns:
            if column not in update_cols:
                updateData.drop_in_place(column)

        iters = (
            {
                "type": self.source_configuration.datamodel,
                "data": x,
                "comment": "Updated via API Sync",
            }
            for x in updateData.iter_rows(named=True)
        )

        with alive_bar(len(data)) as pr_bar:
            if len(data) > BULK_CUTOFF:
                with RequestThreader(50) as needle:
                    for _ in needle.submitAndWait(self.update_item, iters):
                        pr_bar()
            else:
                for item in iters:
                    self.update_item(**item)
                    pr_bar()

        return len(data)

    def delete(self, data: pl.DataFrame):
        deleteData = data.clone()
        deleteFunc = self.delete_item
        deleteMessage = "Deleted via API Sync"

        if self.soft_delete_active():
            soft_delete = self.source_configuration.soft_delete

            deleteData = deleteData.with_columns(
                pl.Series(
                    soft_delete.get("field", "status"),
                    [soft_delete.get("inactive_value", "inactive")] * len(deleteData),
                )
            )
            deleteFunc = self.update_item
            deleteMessage = "Soft deletion via API Sync"

        for column_name in deleteData.columns:
            if column_name in self.source_configuration.link_columns:
                deleteData.drop_in_place(column_name)

        iters = (
            {
                "type": self.source_configuration.datamodel,
                "data": x,
                "comment": deleteMessage,
            }
            for x in deleteData.iter_rows(named=True)
        )

        with alive_bar(len(data)) as pr_bar:
            if len(data) > BULK_CUTOFF:
                with RequestThreader(100) as needle:
                    for _ in needle.submitAndWait(deleteFunc, iters):
                        pr_bar()
            else:
                for item in iters:
                    deleteFunc(**item)
                    pr_bar()

        return len(data)
